@@ -15,19 +15,20 @@
  *  limitations under the License.
  *
  */
-use std::io;
-use std::os::fd::{FromRawFd, OwnedFd};
-use std::time::Duration;
-
-use anyhow::{anyhow, Result};
-use app::App;
+use crate::helpers::format_percent;
+use anyhow::{anyhow, Context, Result};
+use app::SortColumn;
+use app::{App, Mode};
 use bpf_program::BpfProgram;
 use crossterm::event::{self, poll, Event, KeyCode, KeyModifiers};
 use crossterm::execute;
 use crossterm::terminal::{
     disable_raw_mode, enable_raw_mode, EnterAlternateScreen, LeaveAlternateScreen,
 };
+use libbpf_rs::skel::{OpenSkel, Skel, SkelBuilder};
 use libbpf_sys::bpf_enable_stats;
+use pid_iter::PidIterSkelBuilder;
+use procfs::KernelVersion;
 use ratatui::backend::{Backend, CrosstermBackend};
 use ratatui::layout::{Constraint, Direction, Layout, Rect};
 use ratatui::style::{Color, Modifier, Style, Stylize};
@@ -37,21 +38,35 @@ use ratatui::widgets::{
     Table,
 };
 use ratatui::{symbols, Frame, Terminal};
+use std::fs;
+use std::io::{self, Stdout};
+use std::os::fd::{FromRawFd, OwnedFd};
+use std::panic;
+use std::time::Duration;
+use tracing::info;
+use tracing_subscriber::layer::SubscriberExt;
+use tracing_subscriber::util::SubscriberInitExt;
+use tui_input::backend::crossterm::EventHandler;
 
 mod app;
 mod bpf_program;
+mod helpers;
+mod pid_iter {
+    include!(concat!(
+        env!("CARGO_MANIFEST_DIR"),
+        "/src/bpf/pid_iter.skel.rs"
+    ));
+}
 
-const TABLE_FOOTER: &str = "(q) quit | (↑,k) move up | (↓,j) move down | (↵) show graphs";
+const TABLE_FOOTER: &str =
+    "(q) quit | (↑,k) move up | (↓,j) move down | (↵) show graphs | (f) filter | (s) sort";
 const GRAPHS_FOOTER: &str = "(q) quit | (↵) show program list";
-const HEADER_COLS: [&str; 7] = [
-    "ID",
-    "Type",
-    "Name",
-    "Period Avg Runtime (ns)",
-    "Total Avg Runtime (ns)",
-    "Events per second",
-    "Total CPU %",
-];
+const FILTER_FOOTER: &str = "(↵,Esc) back";
+const SORT_CONTROLS_FOOTER: &str =
+    "(↑) asc | (↓) desc | (Backspace) clear | (←) move left | (→) move right";
+const SORT_INFO_FOOTER: &str = "(Esc) back";
+
+const PROCFS_BPF_STATS_ENABLED: &str = "/proc/sys/kernel/bpf_stats_enabled";
 
 impl From<&BpfProgram> for Row<'_> {
     fn from(bpf_program: &BpfProgram) -> Self {
@@ -70,38 +85,108 @@ impl From<&BpfProgram> for Row<'_> {
     }
 }
 
+/// Responsible for managing the terminal state and cleaning up when the program exits
+struct TerminalManager {
+    terminal: Terminal<CrosstermBackend<Stdout>>,
+}
+
+impl TerminalManager {
+    fn new() -> Result<Self> {
+        enable_raw_mode()?;
+        let mut stdout = io::stdout();
+        execute!(stdout, EnterAlternateScreen)?;
+        let backend = CrosstermBackend::new(stdout);
+        let mut terminal = Terminal::new(backend)?;
+        terminal.hide_cursor()?;
+        Ok(Self { terminal })
+    }
+}
+
+impl Drop for TerminalManager {
+    fn drop(&mut self) {
+        execute!(self.terminal.backend_mut(), LeaveAlternateScreen)
+            .unwrap_or_else(|e| eprintln!("Error leaving alternate screen: {:?}", e));
+        disable_raw_mode().unwrap_or_else(|e| eprintln!("Error disabling raw mode: {:?}", e));
+        self.terminal
+            .show_cursor()
+            .unwrap_or_else(|e| eprintln!("Error showing cursor: {:?}", e));
+    }
+}
+
 fn main() -> Result<()> {
-    if !running_as_root() {
+    if !nix::unistd::Uid::current().is_root() {
         return Err(anyhow!("This program must be run as root"));
     }
 
-    // Try to enable BPF stats
-    let fd = unsafe { bpf_enable_stats(libbpf_sys::BPF_STATS_RUN_TIME) };
-    if fd < 0 {
-        return Err(anyhow!("Failed to enable BPF stats"));
+    // Initialize the tracing subscriber with the journald layer
+    let registry = tracing_subscriber::registry()
+        .with(tracing_journald::layer()?)
+        .with(tracing_subscriber::filter::LevelFilter::INFO);
+    // Try to set this subscriber as the global default
+    registry.try_init()?;
+
+    let kernel_version = KernelVersion::current()?;
+    let _owned_fd: OwnedFd;
+    let mut stats_enabled_via_procfs = false;
+    let mut iter_link = None;
+
+    info!("Starting bpftop...");
+    info!("Kernel: {:?}", kernel_version);
+
+    // enable BPF stats via syscall if kernel version >= 5.8
+    if kernel_version >= KernelVersion::new(5, 8, 0) {
+        let fd = unsafe { bpf_enable_stats(libbpf_sys::BPF_STATS_RUN_TIME) };
+        if fd < 0 {
+            return Err(anyhow!("Failed to enable BPF stats via syscall"));
+        }
+        _owned_fd = unsafe { OwnedFd::from_raw_fd(fd) };
+        info!("Enabled BPF stats via syscall");
+
+        // load and attach pid_iter BPF program to get process information
+        let skel_builder = PidIterSkelBuilder::default();
+        let open_skel = skel_builder.open()?;
+        let mut skel = open_skel.load()?;
+        skel.attach()?;
+        iter_link = skel.links.bpftop_iter;
+    } else {
+        // otherwise, enable via procfs
+        // but first check if procfs bpf stats were already enabled
+        if procfs_bpf_stats_is_enabled()? {
+            info!("BPF stats already enabled via procfs");
+        } else {
+            fs::write(PROCFS_BPF_STATS_ENABLED, b"1").context(format!(
+                "Failed to enable BPF stats via {}",
+                PROCFS_BPF_STATS_ENABLED
+            ))?;
+            stats_enabled_via_procfs = true;
+            info!("Enabled BPF stats via procfs");
+        }
     }
-    // The file descriptor will be closed when `_owned_fd` goes out of scope.
-    // This guarantees that BPF stats will be disabled when the program exits.
-    let _owned_fd = unsafe { OwnedFd::from_raw_fd(fd) };
+
+    // capture panic to disable BPF stats via procfs
+    let previous_hook = panic::take_hook();
+    panic::set_hook(Box::new(move |panic_info| {
+        if stats_enabled_via_procfs {
+            if let Err(err) = procs_bfs_stats_disable() {
+                eprintln!("Failed to disable BPF stats via procfs: {:?}", err);
+            }
+        }
+
+        previous_hook(panic_info);
+    }));
 
     // setup terminal
-    enable_raw_mode()?;
-    let mut stdout = io::stdout();
-    execute!(stdout, EnterAlternateScreen)?;
-    let backend = CrosstermBackend::new(stdout);
-    let mut terminal = Terminal::new(backend)?;
-    terminal.clear()?;
+    let mut terminal_manager = TerminalManager::new()?;
 
     // create app and run the draw loop
     let app = App::new();
-    app.start_background_thread();
-    let res = run_draw_loop(&mut terminal, app);
+    app.start_background_thread(iter_link);
+    let res = run_draw_loop(&mut terminal_manager.terminal, app);
 
-    // restore terminal
-    disable_raw_mode()?;
-    execute!(terminal.backend_mut(), LeaveAlternateScreen,)?;
-    terminal.show_cursor()?;
-    terminal.clear()?;
+    // disable BPF stats via procfs if needed
+    if stats_enabled_via_procfs {
+        procs_bfs_stats_disable()?;
+    }
 
     #[allow(clippy::question_mark)]
     if res.is_err() {
@@ -111,6 +196,20 @@ fn main() -> Result<()> {
     Ok(())
 }
 
+fn procs_bfs_stats_disable() -> Result<()> {
+    fs::write(PROCFS_BPF_STATS_ENABLED, b"0").context(format!(
+        "Failed to disable BPF stats via {}",
+        PROCFS_BPF_STATS_ENABLED
+    ))?;
+    Ok(())
+}
+
+fn procfs_bpf_stats_is_enabled() -> Result<bool> {
+    fs::read_to_string(PROCFS_BPF_STATS_ENABLED)
+        .context(format!("Failed to read from {}", PROCFS_BPF_STATS_ENABLED))
+        .map(|value| value.trim() == "1")
+}
+
 fn run_draw_loop<B: Backend>(terminal: &mut Terminal<B>, mut app: App) -> Result<()> {
     loop {
         terminal.draw(|f| ui(f, &mut app))?;
@@ -118,30 +217,47 @@ fn run_draw_loop<B: Backend>(terminal: &mut Terminal<B>, mut app: App) -> Result
         // wait up to 100ms for a keyboard event
         if poll(Duration::from_millis(50))? {
             if let Event::Key(key) = event::read()? {
-                match key.code {
-                    KeyCode::Down | KeyCode::Char('j')=> {
-                        if !app.show_graphs {
-                            app.next_program()
+                match app.mode {
+                    Mode::Table => match key.code {
+                        KeyCode::Down | KeyCode::Char('j') => app.next_program(),
+                        KeyCode::Up | KeyCode::Char('k') => app.previous_program(),
+                        KeyCode::Enter => app.show_graphs(),
+                        KeyCode::Char('f') => app.toggle_filter(),
+                        KeyCode::Char('s') => app.toggle_sort(),
+                        KeyCode::Char('q') | KeyCode::Esc => return Ok(()),
+                        _ => {}
+                    },
+                    Mode::Graph => match key.code {
+                        KeyCode::Enter | KeyCode::Esc => app.show_table(),
+                        KeyCode::Char('q') => return Ok(()),
+                        _ => {}
+                    },
+                    Mode::Filter => match key.code {
+                        KeyCode::Enter | KeyCode::Esc => app.toggle_filter(),
+                        _ => {
+                            app.filter_input
+                                .lock()
+                                .unwrap()
+                                .handle_event(&Event::Key(key));
                         }
-                    }
-                    KeyCode::Up | KeyCode::Char('k') => {
-                        if !app.show_graphs {
-                            app.previous_program()
-                        }
-                    }
-                    KeyCode::Enter => app.toggle_graphs(),
-                    KeyCode::Char('q') => return Ok(()),
-                    KeyCode::Esc => {
-                        if app.show_graphs {
-                            app.toggle_graphs()
-                        } else {
-                            return Ok(());
-                        }
-                    }
-                    _ => {}
+                    },
+                    Mode::Sort => match key.code {
+                        KeyCode::Esc => app.toggle_sort(),
+                        KeyCode::Up => app.sort_column(SortColumn::Ascending(
+                            app.selected_column.unwrap_or_default(),
+                        )),
+                        KeyCode::Down => app.sort_column(SortColumn::Descending(
+                            app.selected_column.unwrap_or_default(),
+                        )),
+                        KeyCode::Backspace => app.sort_column(SortColumn::NoOrder),
+                        KeyCode::Left => app.previous_column(),
+                        KeyCode::Right => app.next_column(),
+                        KeyCode::Enter => app.cycle_sort_exit(),
+                        _ => {}
+                    },
                 }
                 if let (KeyModifiers::CONTROL, KeyCode::Char('c')) = (key.modifiers, key.code) {
-                    return Ok(())
+                    return Ok(());
                 }
             }
         }
@@ -151,16 +267,9 @@ fn run_draw_loop<B: Backend>(terminal: &mut Terminal<B>, mut app: App) -> Result
 fn ui(f: &mut Frame, app: &mut App) {
     let rects = Layout::vertical([Constraint::Min(5), Constraint::Length(3)]).split(f.size());
 
-    // This can happen when the program exists while the user is viewing the graphs.
-    // In this case, we want to switch back to the table view.
-    if app.selected_program().is_none() && app.show_graphs {
-        app.show_graphs = false;
-    }
-
-    if app.show_graphs {
-        render_graphs(f, app, rects[0]);
-    } else {
-        render_table(f, app, rects[0]);
+    match app.mode {
+        Mode::Table | Mode::Filter | Mode::Sort => render_table(f, app, rects[0]),
+        Mode::Graph => render_graphs(f, app, rects[0]),
     }
     render_footer(f, app, rects[1]);
 }
@@ -213,9 +322,15 @@ fn render_graphs(f: &mut Frame, app: &mut App, area: Rect) {
     let max_cpu = moving_max_cpu;
     let max_eps = moving_max_eps as f64;
     let max_runtime = moving_max_runtime as f64;
-    let avg_cpu = total_cpu / data_buf.len() as f64;
-    let avg_eps = total_eps as f64 / data_buf.len() as f64;
-    let avg_runtime = total_runtime as f64 / data_buf.len() as f64;
+
+    let mut avg_cpu = 0.0;
+    let mut avg_eps = 0.0;
+    let mut avg_runtime = 0.0;
+    if data_buf.len() > 0 {
+        avg_cpu = total_cpu / data_buf.len() as f64;
+        avg_eps = total_eps as f64 / data_buf.len() as f64;
+        avg_runtime = total_runtime as f64 / data_buf.len() as f64;
+    }
 
     let cpu_y_max = app.max_cpu.ceil();
     let eps_y_max = (app.max_eps as f64 * 2.0).ceil();
@@ -243,7 +358,7 @@ fn render_graphs(f: &mut Frame, app: &mut App, area: Rect) {
         .block(
             Block::default()
                 .title(format!(
-                    " Total CPU % | Moving Avg: {} | Max: {}",
+                    " Total CPU % | Moving Avg: {} | Max: {} ",
                     format_percent(avg_cpu),
                     format_percent(max_cpu)
                 ))
@@ -274,7 +389,7 @@ fn render_graphs(f: &mut Frame, app: &mut App, area: Rect) {
         .block(
             Block::default()
                 .title(format!(
-                    " Events per second | Moving Avg: {} | Max: {}",
+                    " Events per second | Moving Avg: {} | Max: {} ",
                     avg_eps.ceil(),
                     max_eps.ceil()
                 ))
@@ -305,7 +420,7 @@ fn render_graphs(f: &mut Frame, app: &mut App, area: Rect) {
         .block(
             Block::default()
                 .title(format!(
-                    " Avg Runtime (ns) | Moving Avg: {} | Max: {}",
+                    " Avg Runtime (ns) | Moving Avg: {} | Max: {} ",
                     avg_runtime.ceil(),
                     max_runtime.ceil()
                 ))
@@ -336,20 +451,35 @@ fn render_graphs(f: &mut Frame, app: &mut App, area: Rect) {
     ];
     let widths = [Constraint::Length(15), Constraint::Min(0)];
 
-    if let Some(bpf_program) = app.selected_program() {
+    if let Some(bpf_program) = app.graphs_bpf_program.lock().unwrap().clone() {
         items = vec![
             Row::new(vec![
                 Cell::from("Program ID".bold()),
-                Cell::from(bpf_program.id),
-            ]).height(2),
+                Cell::from(bpf_program.id.to_string()),
+            ])
+            .height(2),
             Row::new(vec![
                 Cell::from("Program Type".bold()),
                 Cell::from(bpf_program.bpf_type),
-            ]).height(2),
+            ])
+            .height(2),
             Row::new(vec![
                 Cell::from("Program Name".bold()),
                 Cell::from(bpf_program.name),
-            ]).height(2),
+            ])
+            .height(2),
+            Row::new(vec![
+                Cell::from("PIDs".bold()),
+                Cell::from(
+                    bpf_program
+                        .processes
+                        .iter()
+                        .map(|pid| pid.to_string())
+                        .collect::<Vec<String>>()
+                        .join(", "),
+                ),
+            ])
+            .height(2),
         ];
     }
 
@@ -371,7 +501,22 @@ fn render_graphs(f: &mut Frame, app: &mut App, area: Rect) {
 fn render_table(f: &mut Frame, app: &mut App, area: Rect) {
     let selected_style = Style::default().add_modifier(Modifier::REVERSED);
     let normal_style = Style::default().bg(Color::Blue);
-    let header = Row::new(HEADER_COLS)
+
+    let columns: Vec<Cell<'_>> = app
+        .header_columns
+        .iter()
+        .enumerate()
+        .map(|(i, col)| {
+            Cell::new(&**col).style(
+                if app.selected_column.is_some_and(|selected| selected == i) {
+                    selected_style
+                } else {
+                    normal_style
+                },
+            )
+        })
+        .collect();
+    let header = Row::new(columns)
         .style(normal_style)
         .height(1)
         .bottom_margin(1);
@@ -399,55 +544,70 @@ fn render_table(f: &mut Frame, app: &mut App, area: Rect) {
         )
         .highlight_style(selected_style)
         .highlight_symbol(">> ");
-    f.render_stateful_widget(t, area, &mut app.state.lock().unwrap());
+    f.render_stateful_widget(t, area, &mut app.table_state);
 }
 
 fn render_footer(f: &mut Frame, app: &mut App, area: Rect) {
-    let footer_text = if app.show_graphs {
-        GRAPHS_FOOTER
-    } else {
-        TABLE_FOOTER
+    let info_text = match app.mode {
+        Mode::Table => TABLE_FOOTER,
+        Mode::Graph => GRAPHS_FOOTER,
+        Mode::Filter => FILTER_FOOTER,
+        Mode::Sort => SORT_INFO_FOOTER,
     };
-    let info_footer = Paragraph::new(Line::from(footer_text)).centered().block(
+    let info_footer = Paragraph::new(Line::from(info_text)).centered().block(
         Block::default()
             .borders(Borders::ALL)
             .border_type(BorderType::Double),
     );
-    f.render_widget(info_footer, area);
-}
 
-fn running_as_root() -> bool {
-    matches!(nix::unistd::getuid().as_raw(), 0)
-}
-
-fn format_percent(num: f64) -> String {
-    if num < 1.0 {
-        round_to_first_non_zero(num).to_string() + "%"
-    } else {
-        format!("{:.2}%", num)
-    }
-}
-
-fn round_to_first_non_zero(num: f64) -> f64 {
-    if num == 0.0 {
-        return 0.0;
+    // Only single footer in table and graph mode
+    if let Mode::Table | Mode::Graph = app.mode {
+        f.render_widget(info_footer, area);
+        return;
     }
 
-    let mut multiplier = 1.0;
-    while num * multiplier < 1.0 {
-        multiplier *= 10.0;
-    }
-    (num * multiplier).round() / multiplier
-}
+    // Two footers in filter and sort mode
+    let split_area = Layout::default()
+        .direction(Direction::Horizontal)
+        .constraints([Constraint::Percentage(75), Constraint::Percentage(25)].as_ref())
+        .split(area);
 
-#[cfg(test)]
-mod tests {
-    use super::*;
+    // Right footer
+    f.render_widget(info_footer, split_area[1]);
 
-    #[test]
-    fn test_round_to_first_non_zero() {
-        assert_eq!(round_to_first_non_zero(0.002323), 0.002);
-        assert_eq!(round_to_first_non_zero(0.0000012), 0.000001);
-        assert_eq!(round_to_first_non_zero(0.00321), 0.003);
+    // Left footer
+    match app.mode {
+        Mode::Filter => {
+            let filter_input = app.filter_input.lock().unwrap();
+            let filter_footer = Paragraph::new(filter_input.value()).block(
+                Block::default()
+                    .padding(Padding::horizontal(1))
+                    .borders(Borders::ALL)
+                    .border_type(BorderType::Double)
+                    .title(" Filter Name/Type "),
+            );
+
+            f.render_widget(filter_footer, split_area[0]);
+
+            // Displays cursor when inputting
+            f.set_cursor(
+                split_area[0].x + filter_input.visual_cursor() as u16 + 2,
+                split_area[0].y + 1,
+            );
+            drop(filter_input);
+        }
+        Mode::Sort => {
+            let sort_footer = Paragraph::new(Line::from(SORT_CONTROLS_FOOTER))
+                .centered()
+                .block(
+                    Block::default()
+                        .borders(Borders::ALL)
+                        .border_type(BorderType::Double)
+                        .title(" Sort Column "),
+                );
+
+            f.render_widget(sort_footer, split_area[0]);
+        }
+        _ => {}
     }
 }
